package edu.uiuc.ncsa.security.util.cli;

import edu.uiuc.ncsa.security.core.Logable;
import edu.uiuc.ncsa.security.core.util.LoggerProvider;
import edu.uiuc.ncsa.security.core.util.MyLoggingFacade;
import org.apache.commons.cli.*;

/**
 * Basic class for writing command line tools. This allows for verbose, debug and help options automatically.
 * It supports logging to a file and specifying a configuration file names. In short, a lot
 * of the random grunt work for using the commons command line toolkit is here.
 * <p>Created by Jeff Gaynor<br>
 * on 9/7/11 at  12:31 PM
 */
public abstract class CLITool implements Logable {
    public static final String VERBOSE_OPTION = "v"; // if present do a backup of the target to a backup schema.
    public static final String VERBOSE_LONG_OPTION = "verbose"; // if present do a backup of the target to a backup schema.

    public static final String DEBUG_OPTION = "d"; // if present do a backup of the target to a backup schema.
    public static final String DEBUG_LONG_OPTION = "debug"; // if present do a backup of the target to a backup schema.


    public static final String LOG_FILE_OPTION = "log"; // if present do a backup of the target to a backup schema.
    public static final String LOG_FILE_LONG_OPTION = "logFile"; // if present do a backup of the target to a backup schema.

    // other options
    public static final String HELP_OPTION = "h";
    public static final String HELP_LONG_OPTION = "help";


    public static final String CONFIG_FILE_OPTION = "cfg";
    public static final String CONFIG_FILE_LONG_OPTION = "configFile";


    public static final String DEFAULT_LOG_FILE = "log.xml";
    /**
     * The default action for this tool at startup. Do nothing.
     */
    public static final int NO_ACTION = 0;

    boolean verbose = false;
    boolean debugOn = true;

    String logfileName = DEFAULT_LOG_FILE;

    /**
     * Does the actual work. This is where you put your executable code.
     */
    public abstract void doIt() throws Exception;

    /**
     * Sets up the configuration and runtime environment.
     */
    public abstract void initialize() throws Exception;


    /**
     * This will be automatically invoked for you if the user supplies a help option. This should
     * just print out to the command line (using the {@link #say(String)} method), rather than logging
     * the help.
     */
    public abstract void help();

    /**
     * Convenience method. This prints out a short message describing the built-in options for this
     * class. Best called at the end of your custom help method.
     * Prints a header (informational line) if you wish.
     */
    public void defaultHelp(boolean printHeader) {
        if (printHeader) {
            say("Standard options are:");
        }
        say("  -" + LOG_FILE_OPTION + " (-" + LOG_FILE_LONG_OPTION + ") -- set the name of the log file. Default is \"" + DEFAULT_LOG_FILE + "\" in the invocation directory.");
        say("  -" + CONFIG_FILE_OPTION + " (-" + CONFIG_FILE_LONG_OPTION + ") -- the configuration file for this application.");
        say("  -" + VERBOSE_OPTION + " (-" + VERBOSE_LONG_OPTION + ") -- set verbose on. This sends output to the console.");
        say("  -" + DEBUG_OPTION + " (-" + DEBUG_LONG_OPTION + ") -- set debug on. This will print out extra messages \n\t\t\tduring execution to the log file (and console, if verbose is enabled)");
        say("  -" + HELP_OPTION + " (-" + HELP_LONG_OPTION + ") -- this message");
    }

    public String getConfigFile() {
        return configFile;
    }

    public void setConfigFile(String configFile) {
        this.configFile = configFile;
    }

    String configFile;


    public String getLogfileName() {
        return logfileName;
    }

    public void setLogfileName(String logfileName) {
        this.logfileName = logfileName;
    }

    public boolean isVerbose() {
        return verbose;
    }

    public void setVerbose(boolean verbose) {
        this.verbose = verbose;
    }


    /**
     * Returns true if execution should continue, false if not. Call this first to see if execution should proceed.
     *
     * @param args
     * @return
     * @throws Exception
     */
    protected boolean getOptions(String[] args) throws Exception {
        getOptions();
        if (args.length == 0) {
            help();
            return false;
        }
        for (String z : args) {
            if (z.toLowerCase().endsWith("help")) {
                help();
                return false;
            }
        }
        parseCommandLine(args);
        if (hasOption(HELP_OPTION, HELP_LONG_OPTION)) {
            help();
            return false;
        }
        setVerbose(false);

        if (hasOption(VERBOSE_OPTION, VERBOSE_LONG_OPTION)) {
            setVerbose(true);
        }

        setDebugOn(false);
        if (hasOption(DEBUG_OPTION, DEBUG_LONG_OPTION)) {
            setDebugOn(true);
        }

        if (hasOption(LOG_FILE_OPTION, LOG_FILE_LONG_OPTION)) {
            setLogfileName(getCommandLine().getOptionValue(LOG_FILE_LONG_OPTION));
        }

        if (hasOption(CONFIG_FILE_OPTION, CONFIG_FILE_LONG_OPTION)) {
            setConfigFile(getCommandLine().getOptionValue(CONFIG_FILE_OPTION));
        }
        return true;
    }

    /**
     * Checks if the long or short form is part of the command line options.
     *
     * @param shortForm
     * @param longForm
     * @return
     */
    protected boolean hasOption(String shortForm, String longForm) {
        return getCommandLine().hasOption(shortForm) || getCommandLine().hasOption(longForm);
    }

    CommandLine cmd = null;

    protected void parseCommandLine(String[] args) throws UnrecognizedOptionException, ParseException {
        CommandLineParser clp = new BasicParser();
        try {
            cmd = clp.parse(getOptions(), args);
        } catch (UnrecognizedOptionException ux) {
            say("Error: unrecognized option  + " + ux.getMessage());
            say("Invoke with -help for more");
            if (isVerbose()) {
                ux.printStackTrace();
            }
            throw ux;
        } catch (ParseException e) {
            say("Error: could not parse a command line argument:" + e.getMessage());
            if (isVerbose()) {
                e.printStackTrace();
            }
            throw e;
        }
    }

    public CommandLine getCommandLine() {
        return cmd;
    }

    /**
     * What this tool should do.
     */
    int action;

    /**
     * This application allows you to specify numeric values that determine the behavior of this application.
     * You may access the result in a switch statement. You set them in the {@link #getOptions(String[])}
     * method and access them in the {@link #doIt()} method. Use of this facility, is optional.
     *
     * @return
     */
    public int getAction() {
        return action;
    }

    public void setAction(int action) {
        this.action = action;
    }


    /**
     * Converts command line arguments into integer action values. This makes writing switch
     * statements for choosing actions very easy. Pass in the short and long options, the
     * action that should be assigned and the command line. If the command option is present,
     * the action will be set.
     * <p/>
     * <p>The only supplied default is {@link #NO_ACTION} </p>
     *
     * @param flag
     * @param longFlag
     * @param action
     */
    protected void checkAction(String flag, String longFlag, int action) {
        if (getCommandLine().hasOption(flag) || getCommandLine().hasOption(longFlag)) {
            if (getAction() != NO_ACTION) {
                throw new RuntimeException("Error: attempted to reset the action");
            }
            setAction(action);
        }

    }

    /**
     * Override this to set up your options. You should also use this to check the action and
     * set it, .e.g.,<br><br>
     * <code>
     * &nbsp;&nbsp;&nbsp;&nbsp;Options options = super.getOptions();<br>
     * &nbsp;&nbsp;&nbsp;&nbsp;options.addOptions(SHORT_FORM, LONG_FORM, ... // This is in the commons documentation <BR>
     * &nbsp;&nbsp;&nbsp;&nbsp;checkAction(SHORT_FORM, LONG_FORM, ACTION_CODE);<BR>
     * </code>
     * <br><br>
     * which would set an integer value suitable for using with the {@link #getAction()} call in
     * a switch statement.
     *
     * @return
     */
    protected Options getOptions() {
        Options options = new Options();
        options.addOption(HELP_OPTION, HELP_LONG_OPTION, false, "Display the help message.");
        options.addOption(DEBUG_OPTION, DEBUG_LONG_OPTION, false, "Enable/disable debug mode.");
        options.addOption(VERBOSE_OPTION, VERBOSE_LONG_OPTION, false, "Set verbose mode on");
        options.addOption(CONFIG_FILE_OPTION, CONFIG_FILE_LONG_OPTION, true, "Set the configuration file");
        options.addOption(LOG_FILE_OPTION, LOG_FILE_LONG_OPTION, true, "Set the log file");
        return options;
    }

    /**
     * Prints to the console only if verbose is enabled.
     *
     * @param x
     */
    protected void sayv(String x) {
            info(x);
    }

    /**
     * Prints to the console.
     *
     * @param x
     */
    public void say(String x) {
        System.out.println(x);
    }

    public void setMyLogger(MyLoggingFacade  myLoggingFacade){
        logger = myLoggingFacade;

    }
    public MyLoggingFacade getMyLogger() {

        if (logger == null) {
            LoggerProvider loggerProvider = new LoggerProvider(getLogfileName(), "cli logger", 1,1000000,isDebugOn(), true);
            logger = loggerProvider.get();
        }
        return logger;
    }

    MyLoggingFacade logger;

    public void debug(String x) {
        if (isVerbose() && isDebugOn()) {
            say(x);
        }
        getMyLogger().debug(x);
    }

    public boolean isDebugOn() {
        return debugOn;
    }

    public void setDebugOn(boolean setOn) {
        this.debugOn = setOn;
    }

    public void info(String x) {
        if (isVerbose()) {
            say(x);
        }
        getMyLogger().info(x);
    }

    public void warn(String x) {
        if (isVerbose()) {
            say(x);
        }
        getMyLogger().warn(x);
    }

    public void error(String x) {
        if (isVerbose()) {
            say(x);
        }
        getMyLogger().error(x);
    }

    /**
     * Main call. This will grab the command line arguments, run {@link #initialize()}
     * and then invoke {@link #doIt()}.
     * @param args
     * @throws Throwable
     */
    public void run(String[] args) throws Throwable {
        if (!getOptions(args)) {
            return;
        }
        initialize();
        doIt();
    }


}
