package edu.uiuc.ncsa.security.util.pkcs;

/**
 * <p>Created by Jeff Gaynor<br>
 * on Feb 23, 2011 at  11:45:07 AM
 */

import edu.uiuc.ncsa.security.core.exceptions.GeneralException;
import org.apache.commons.codec.binary.Base64;
import org.bouncycastle.jce.PKCS10CertificationRequest;

import javax.security.auth.x500.X500Principal;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.*;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Collection;
import java.util.logging.Logger;

/**
 * Utilities for formatting certs and other things.
 * <p>Created by Jeff Gaynor<br>
 * on Jun 15, 2010 at  2:45:47 PM
 */
public class CertUtil {
    public static final String UTF_8 = "UTF-8"; // encoding for all of this.

    static Logger logger;
    /**
     * This required by the specification
     */
    public static final String BEGIN_CERTIFICATE = "-----BEGIN CERTIFICATE-----";
    public static final String END_CERTIFICATE = "-----END CERTIFICATE-----";
    public static final String DEFAULT_PKCS10_SIGNATURE_ALGORITHM = "SHA1withRSA";
    public static final String DEFAULT_PKCS10_PROVIDER = "SunRsaSign";
    public static final String DEFAULT_PKCS10_DISTINGUISHED_NAME = "CN=ignore";


    public static Logger getLogger() {
        if (logger == null) {
            logger = Logger.getLogger(CertUtil.class.getName());
        }
        return logger;
    }


    /**
     * A utility that will take a list of Base64 encoded X509 certificates and convert them into
     * an array of java.security.cert.X509Certificate objects.
     *
     * @param certList
     * @param nameList The array of file names corresponding to each cert. This is only used to generate error messages
     *                 and if it is omitted will not otherwise effect function.
     * @return
     * @throws CertificateException
     */
    public static X509Certificate[] getX509CertsFromStringList(
            String[] certList, String[] nameList) throws CertificateException {
        Collection<X509Certificate> c = new ArrayList<X509Certificate>(
                certList.length);
        for (int i = 0; i < certList.length; i++) {
            int index = -1;
            String certData = certList[i];
            if (certData != null) {
                index = certData.indexOf(BEGIN_CERTIFICATE);
            }
            if (index >= 0) {
                certData = certData.substring(index);
                ByteArrayInputStream inputStream = new ByteArrayInputStream(
                        certData.getBytes());
                try {
                    X509Certificate cert = (X509Certificate) getCertFactory()
                            .generateCertificate(inputStream);
                    c.add(cert);
                } catch (Exception e) {
                    if (nameList != null) {
                        getLogger().warning(nameList[i]
                                + " can not be parsed as an X509Certificate.");
                    } else {
                        getLogger().warning("failed to parse an X509Certificate");
                    }
                }
            }
        }
        if (c.isEmpty())
            return null;
        return c.toArray(new X509Certificate[0]);
    }

    /**
     * Take an X509 cert and encode it correctly in PEM format, writing the result to the
     * output stream. This will check for a known condition in various security suites (e.g. Open SSL)
     * in which an extra line feed will cause the entire result to be rejected. This extra line feed
     * occurs in the odd case that the encoded cert's length is a multiple of the line length. The only
     * way to catch this is to encode it as a string, then check if the last character is a line feed.
     *
     * @param x509Certificate
     * @param outputStream
     * @throws java.security.cert.CertificateEncodingException
     *
     */
    public static void toPEM(X509Certificate x509Certificate, OutputStream outputStream) throws CertificateEncodingException {
        PEMFormatUtil.delimitBody(x509Certificate.getEncoded(), BEGIN_CERTIFICATE, END_CERTIFICATE, outputStream);
    }


    /**
     * Take a string that is an base64-encoded, DER-encoded PKCS10 certificate request,. This is, for instance
     * stored in a database and not PEM format.
     */

    public static PKCS10CertificationRequest fromStringToCertReq(String x) {
        return new PKCS10CertificationRequest(Base64.decodeBase64(x));
    }

    public static String fromCertReqToString(PKCS10CertificationRequest certReq) {
        return Base64.encodeBase64String(certReq.getEncoded());
    }

    /**
     * Convert a string containing the PEM-format cert to a certificate. Useful if the cert has been
     * persisted, e.g., with a database as a string.
     *
     * @param certString
     * @return
     * @throws CertificateException
     */
    public static X509Certificate[] fromX509PEM(String certString) throws CertificateException {
        try {
            ByteArrayInputStream bs = new ByteArrayInputStream(certString.getBytes(UTF_8));
            return fromPEM(bs);
        } catch (Exception e) {
            throw new GeneralException("Error converting cert string", e);
        }

    }


    /**
     * @param inputStream
     * @return
     * @throws CertificateException
     */
    public static X509Certificate[] fromPEM(InputStream inputStream) throws CertificateException {
        /*
          This strips off the header and tail automatically, so we don't really have to do anything.
         */
        return getCertFactory().generateCertificates(inputStream).toArray(new X509Certificate[0]);
    }

    public static CertificateFactory getCertFactory() throws CertificateException {
        if (certFactory == null) {
            certFactory = CertificateFactory.getInstance("X.509");
        }
        return certFactory;
    }

    public static void setCertFactory(CertificateFactory certFactory) {
        certFactory = certFactory;
    }

    static CertificateFactory certFactory;

    /**
     * Encode to PEM.
     *
     * @param x509Certificate
     * @return
     */
    public static String toPEM(X509Certificate x509Certificate) {
        try {
            // The issue is that getEncoded returns a byte array in DER format, but
            // the inverse operation from the cert factory expects the header and strips it off.
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            toPEM(x509Certificate, baos);
            return new String(baos.toByteArray(), UTF_8);
        } catch (Exception e) {
            throw new GeneralException("Error encoding cert", e);
        }

    }

    public static String toPEM(X509Certificate[] x509Certificates) {
        StringBuffer stringBuffer = new StringBuffer();
        boolean firstPass = true;
        for (X509Certificate x509Certificate : x509Certificates) {
            if (firstPass) {
                stringBuffer.append(toPEM(x509Certificate));
                firstPass = false;
            } else {
                stringBuffer.append("\n" + toPEM(x509Certificate));
            }
        }
        return stringBuffer.toString();
    }

    public static PKCS10CertificationRequest createCertRequest(KeyPair keypair) throws SignatureException, NoSuchProviderException, InvalidKeyException, NoSuchAlgorithmException {
        return createCertRequest(keypair, DEFAULT_PKCS10_SIGNATURE_ALGORITHM, DEFAULT_PKCS10_PROVIDER, DEFAULT_PKCS10_DISTINGUISHED_NAME);
    }

    public static PKCS10CertificationRequest createCertRequest(KeyPair keypair, String sigAlgName, String provider, String dn) throws SignatureException,
            InvalidKeyException, NoSuchProviderException, NoSuchAlgorithmException {
        return new PKCS10CertificationRequest(sigAlgName,
                new X500Principal(dn), keypair.getPublic(), null,
                keypair.getPrivate(), provider);
    }
}
